//this file is part of eMule
//Copyright (C)2002-2008 Merkur ( strEmail.Format("%s@%s", "devteam", "emule-project.net") / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#include "stdafx.h"
#include "emule.h"
#include "QueueListCtrl.h"
#include "OtherFunctions.h"
#include "MenuCmds.h"
#include "ClientDetailDialog.h"
#include "Exceptions.h"
#include "KademliaWnd.h"
#include "emuledlg.h"
#include "FriendList.h"
#include "UploadQueue.h"
#include "UpDownClient.h"
#include "TransferWnd.h"
#include "MemDC.h"
#include "SharedFileList.h"
#include "ClientCredits.h"
#include "PartFile.h"
#include "ChatWnd.h"
#include "Kademlia/Kademlia/Kademlia.h"
#include "Kademlia/Kademlia/Prefs.h"
#include "kademlia/net/KademliaUDPListener.h"
#include "Log.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif


IMPLEMENT_DYNAMIC(CQueueListCtrl, CMuleListCtrl)

BEGIN_MESSAGE_MAP(CQueueListCtrl, CMuleListCtrl)
	ON_NOTIFY_REFLECT(LVN_COLUMNCLICK, OnLvnColumnClick)
	ON_NOTIFY_REFLECT(LVN_GETDISPINFO, OnLvnGetDispInfo)
	ON_NOTIFY_REFLECT(NM_DBLCLK, OnNmDblClk)
	ON_WM_CONTEXTMENU()
	ON_WM_SYSCOLORCHANGE()
END_MESSAGE_MAP()

CQueueListCtrl::CQueueListCtrl()
	: CListCtrlItemWalk(this)
{
	SetGeneralPurposeFind(true);
	SetSkinKey(L"QueuedLv");

	// Barry - Refresh the queue every 10 secs
	VERIFY( (m_hTimer = ::SetTimer(NULL, NULL, 10000, QueueUpdateTimer)) != NULL );
	if (thePrefs.GetVerbose() && !m_hTimer)
		AddDebugLogLine(true,_T("Failed to create 'queue list control' timer - %s"),GetErrorMessage(GetLastError()));
}

CQueueListCtrl::~CQueueListCtrl()
{
	if (m_hTimer)
		VERIFY( ::KillTimer(NULL, m_hTimer) );
}

void CQueueListCtrl::Init()
{
	SetPrefsKey(_T("QueueListCtrl"));
	SetExtendedStyle(LVS_EX_FULLROWSELECT | LVS_EX_INFOTIP);

	InsertColumn(0, GetResString(IDS_QL_USERNAME),	LVCFMT_LEFT, DFLT_CLIENTNAME_COL_WIDTH);
	InsertColumn(1, GetResString(IDS_FILE),			LVCFMT_LEFT, DFLT_FILENAME_COL_WIDTH);
	InsertColumn(2, GetResString(IDS_FILEPRIO),		LVCFMT_LEFT, DFLT_PRIORITY_COL_WIDTH);
	InsertColumn(3, GetResString(IDS_QL_RATING),	LVCFMT_LEFT,  60);
	InsertColumn(4, GetResString(IDS_SCORE),		LVCFMT_LEFT,  60);
	InsertColumn(5, GetResString(IDS_ASKED),		LVCFMT_LEFT,  60);
	InsertColumn(6, GetResString(IDS_LASTSEEN),		LVCFMT_LEFT, 110);
	InsertColumn(7, GetResString(IDS_ENTERQUEUE),	LVCFMT_LEFT, 110);
	InsertColumn(8, GetResString(IDS_BANNED),		LVCFMT_LEFT,  60);
	InsertColumn(9, GetResString(IDS_UPSTATUS),		LVCFMT_LEFT, DFLT_PARTSTATUS_COL_WIDTH);
	InsertColumn(10,GetResString(IDS_SOFTWARE_LABEL),LVCFMT_LEFT,100,10);
	InsertColumn(11,GetResString(IDS_MODVERSION_LABEL),LVCFMT_LEFT,100,11);

	// <CB Mod : QueueListInfo>
#ifdef CB_MOD_EXTRAINFO_ENABLED
	InsertColumn(12,GetResString(IDS_DL_SIZE),LVCFMT_LEFT,100,12);
	InsertColumn(13,GetResString(IDS_IPNUMBER),LVCFMT_LEFT,100,13);
	InsertColumn(14,GetResString(IDS_RANK),LVCFMT_LEFT,60,14);
#endif
	// </CB Mod : QueueListInfo

	SetAllIcons();
	Localize();
	LoadSettings();
	SetSortArrow();
	SortItems(SortProc, GetSortItem() + (GetSortAscending() ? 0 : 100));
}

void CQueueListCtrl::Localize()
{
	CHeaderCtrl *pHeaderCtrl = GetHeaderCtrl();
	HDITEM hdi;
	hdi.mask = HDI_TEXT;

	CString strRes;
	strRes = GetResString(IDS_QL_USERNAME);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(0, &hdi);

	strRes = GetResString(IDS_FILE);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(1, &hdi);

	strRes = GetResString(IDS_FILEPRIO);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(2, &hdi);

	strRes = GetResString(IDS_QL_RATING);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(3, &hdi);

	strRes = GetResString(IDS_SCORE);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(4, &hdi);

	strRes = GetResString(IDS_ASKED);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(5, &hdi);

	strRes = GetResString(IDS_LASTSEEN);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(6, &hdi);

	strRes = GetResString(IDS_ENTERQUEUE);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(7, &hdi);

	strRes = GetResString(IDS_BANNED);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(8, &hdi);
	
	strRes = GetResString(IDS_UPSTATUS);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(9, &hdi);

	strRes = GetResString(IDS_SOFTWARE_LABEL);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(10, &hdi);
	
	strRes = GetResString(IDS_MODVERSION_LABEL);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(11, &hdi);

	// <CB Mod : QueueListInfo>
#ifdef CB_MOD_EXTRAINFO_ENABLED
	strRes = GetResString(IDS_DL_SIZE);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(12, &hdi);

	strRes = GetResString(IDS_IPNUMBER);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(13, &hdi);
	
	strRes = GetResString(IDS_RANK);
	hdi.pszText = const_cast<LPTSTR>((LPCTSTR)strRes);
	pHeaderCtrl->SetItem(14, &hdi);
#endif
	// </CB Mod : QueueListInfo
}


void CQueueListCtrl::OnSysColorChange()
{
	CMuleListCtrl::OnSysColorChange();
	SetAllIcons();
}

void CQueueListCtrl::SetAllIcons()
{
	ApplyImageList(NULL);
	m_ImageList.DeleteImageList();
	m_ImageList.Create(16, 16, theApp.m_iDfltImageListColorFlags | ILC_MASK, 0, 1);
	m_ImageList.Add(CTempIconLoader(_T("ClientEDonkey")));
	m_ImageList.Add(CTempIconLoader(_T("ClientCompatible")));
	m_ImageList.Add(CTempIconLoader(_T("ClientEDonkeyPlus")));
	m_ImageList.Add(CTempIconLoader(_T("ClientCompatiblePlus")));
	m_ImageList.Add(CTempIconLoader(_T("Friend")));
	m_ImageList.Add(CTempIconLoader(_T("ClientMLDonkey")));
	m_ImageList.Add(CTempIconLoader(_T("ClientMLDonkeyPlus")));
	m_ImageList.Add(CTempIconLoader(_T("ClientEDonkeyHybrid")));
	m_ImageList.Add(CTempIconLoader(_T("ClientEDonkeyHybridPlus")));
	m_ImageList.Add(CTempIconLoader(_T("ClientShareaza")));
	m_ImageList.Add(CTempIconLoader(_T("ClientShareazaPlus")));
	m_ImageList.Add(CTempIconLoader(_T("ClientAMule")));
	m_ImageList.Add(CTempIconLoader(_T("ClientAMulePlus")));
	m_ImageList.Add(CTempIconLoader(_T("ClientLPhant")));
	m_ImageList.Add(CTempIconLoader(_T("ClientLPhantPlus")));
	m_ImageList.SetOverlayImage(m_ImageList.Add(CTempIconLoader(_T("ClientSecureOvl"))), 1);
	m_ImageList.SetOverlayImage(m_ImageList.Add(CTempIconLoader(_T("OverlayObfu"))), 2);
	m_ImageList.SetOverlayImage(m_ImageList.Add(CTempIconLoader(_T("OverlaySecureObfu"))), 3);
	// Apply the image list also to the listview control, even if we use our own 'DrawItem'.
	// This is needed to give the listview control a chance to initialize the row height.
	ASSERT( (GetStyle() & LVS_SHAREIMAGELISTS) != 0 );
	VERIFY( ApplyImageList(m_ImageList) == NULL );
}

// <CB Mod : Lister extended background colors : Date 23.03.09>
#ifdef CB_MOD_ENHANCEDINTERFACE_ENABLED
void CQueueListCtrl::InitItemMemDC(CMemDC *dc, LPDRAWITEMSTRUCT lpDrawItemStruct, BOOL &bCtrlFocused) {
	bCtrlFocused = ((GetFocus() == this) || (GetStyle() & LVS_SHOWSELALWAYS));
	COLORREF lineBackColor = m_crWindow;

	const CUpDownClient *client = (CUpDownClient *)lpDrawItemStruct->itemData;
	CKnownFile* pKnownFile = theApp.sharedfiles->GetFileByID(client->GetUploadFileID());
	
	if (pKnownFile != NULL) {
		if (lpDrawItemStruct->itemState & ODS_SELECTED) {
			if (bCtrlFocused) {
				if (client->IsFriend() && client->GetFriendSlot())
					lineBackColor = FRIENDSLOT_BACK_SELECTED;
				else if (pKnownFile->GetPowerShared())
					lineBackColor = POWERSHARED_BACK_SELECTED;
	#ifdef CB_MOD_NEWPRIORITIES_ENABLED
				else if (pKnownFile->IsSuperRelease())
					lineBackColor = SUPERRELEASE_BACK_SELECTED;
	#endif
	#ifdef CB_MOD_SFP_ENABLED
				else if (thePrefs.SmallFilePush() && pKnownFile->IsSFP())
					lineBackColor = SMALLPUSHED_BACK_SELECTED;
	#endif
				else
					lineBackColor = m_crHighlight;
			} else {
				if (client->IsFriend() && client->GetFriendSlot())
					lineBackColor = FRIENDSLOT_BACK_SELECTED_UF;
				else if (pKnownFile->GetPowerShared()) 
					lineBackColor = POWERSHARED_BACK_SELECTED_UF;
	#ifdef CB_MOD_NEWPRIORITIES_ENABLED
				else if (pKnownFile->IsSuperRelease())
					lineBackColor = SUPERRELEASE_BACK_SELECTED_UF;
	#endif
	#ifdef CB_MOD_SFP_ENABLED
				else if (thePrefs.SmallFilePush() && pKnownFile->IsSmallFile())
					lineBackColor = SMALLPUSHED_BACK_SELECTED_UF;
	#endif
			else
				lineBackColor = m_crNoHighlight;

			}
		}
		else
		{
			if (client->IsFriend() && client->GetFriendSlot())
				lineBackColor = FRIENDSLOT_BACK_NORMAL;
			else if (pKnownFile->GetPowerShared())
				lineBackColor = POWERSHARED_BACK_NORMAL;
	#ifdef CB_MOD_NEWPRIORITIES_ENABLED
			else if (pKnownFile->IsSuperRelease())
				lineBackColor = SUPERRELEASE_BACK_NORMAL;
	#endif
	#ifdef CB_MOD_SFP_ENABLED
			else if (thePrefs.SmallFilePush() && pKnownFile->IsSFP())
				lineBackColor = SMALLPUSHED_BACK_NORMAL;
	#endif
			else
				lineBackColor = m_crWindowTextBk;
		}
	} else
		lineBackColor = m_crNoHighlight;

	if (lineBackColor == CLR_NONE)
	{
		DefWindowProc(WM_ERASEBKGND, (WPARAM)(HDC)*dc, 0);
		dc->SetBkMode(TRANSPARENT);
	}
	else
	{
		dc->FillBackground(lineBackColor);
	}
	
	dc->SetTextColor((lpDrawItemStruct->itemState & ODS_SELECTED) ? m_crHighlightText : m_crWindowText);
	dc->SetFont(GetFont());
}
#endif
// </CB Mod : Lister extended background colors>


void CQueueListCtrl::DrawItem(LPDRAWITEMSTRUCT lpDrawItemStruct)
{
	if (!theApp.emuledlg->IsRunning())
		return;
	if (!lpDrawItemStruct->itemData)
		return;

	CMemDC dc(CDC::FromHandle(lpDrawItemStruct->hDC), &lpDrawItemStruct->rcItem);
	BOOL bCtrlFocused;
	InitItemMemDC(dc, lpDrawItemStruct, bCtrlFocused);
	CRect cur_rec(lpDrawItemStruct->rcItem);
	CRect rcClient;
	GetClientRect(&rcClient);
	const CUpDownClient *client = (CUpDownClient *)lpDrawItemStruct->itemData;

	CHeaderCtrl *pHeaderCtrl = GetHeaderCtrl();
	int iCount = pHeaderCtrl->GetItemCount();
	cur_rec.right = cur_rec.left - sm_iLabelOffset;
	cur_rec.left += sm_iIconOffset;

	// <CB Mod : CreditSystem>
	float clientrating = 0;
	// </CB Mod : CreditSystem>

	for (int iCurrent = 0; iCurrent < iCount; iCurrent++)
	{
		int iColumn = pHeaderCtrl->OrderToIndex(iCurrent);
		if (!IsColumnHidden(iColumn))
		{
			UINT uDrawTextAlignment;
			int iColumnWidth = GetColumnWidth(iColumn, uDrawTextAlignment);
			cur_rec.right += iColumnWidth;
			if (cur_rec.left < cur_rec.right && HaveIntersection(rcClient, cur_rec))
			{
				TCHAR szItem[1024];
				GetItemDisplayText(client, iColumn, szItem, _countof(szItem));
				switch (iColumn)
				{
					case 0:{
						int iImage;
					// <CB Mod : CreditSystem>
#ifdef CB_MOD_FINECS_ENABLED
					clientrating = client->credits->GetScoreRatio(client->GetIP(), client);
#else
					clientrating = client->credits->GetScoreRatio(client->GetIP());
#endif
					// </CB Mod : CreditSystem>

						if (client->IsFriend())
							iImage = 4;
						else if (client->GetClientSoft() == SO_EDONKEYHYBRID) {
							// <CB Mod : CreditSystem>
							if ( clientrating > 1)
							// </CB Mod : CreditSystem>
								iImage = 8;
							else
								iImage = 7;
						}
						else if (client->GetClientSoft() == SO_MLDONKEY){
							// <CB Mod : CreditSystem>
							if ( clientrating > 1)
							// </CB Mod : CreditSystem>
								iImage = 6;
							else
								iImage = 5;
						}
						else if (client->GetClientSoft() == SO_SHAREAZA){
							// <CB Mod : CreditSystem>
							if ( clientrating > 1)
							// </CB Mod : CreditSystem>
								iImage = 10;
							else
								iImage = 9;
						}
						else if (client->GetClientSoft() == SO_AMULE){
							// <CB Mod : CreditSystem>
							if ( clientrating > 1)
							// </CB Mod : CreditSystem>
								iImage = 12;
							else
								iImage = 11;
						}
						else if (client->GetClientSoft() == SO_LPHANT){
							// <CB Mod : CreditSystem>
							if ( clientrating > 1)
							// </CB Mod : CreditSystem>
								iImage = 14;
							else
								iImage = 13;
						}
						else if (client->ExtProtocolAvailable()){
							// <CB Mod : CreditSystem>
							if ( clientrating > 1)
							// </CB Mod : CreditSystem>
								iImage = 3;
							else
								iImage = 1;
						}
						else{
							// <CB Mod : CreditSystem>
							if ( clientrating > 1)
							// </CB Mod : CreditSystem>
								iImage = 2;
							else
								iImage = 0;
						}

						UINT nOverlayImage = 0;
						if ((client->Credits() && client->Credits()->GetCurrentIdentState(client->GetIP()) == IS_IDENTIFIED))
							nOverlayImage |= 1;
						if (client->IsObfuscatedConnectionEstablished())
							nOverlayImage |= 2;
						int iIconPosY = (cur_rec.Height() > 16) ? ((cur_rec.Height() - 16) / 2) : 1;
						POINT point = { cur_rec.left, cur_rec.top + iIconPosY };
						m_ImageList.Draw(dc, iImage, point, ILD_NORMAL | INDEXTOOVERLAYMASK(nOverlayImage));

						cur_rec.left += 16 + sm_iLabelOffset;
						dc.DrawText(szItem, -1, &cur_rec, MLC_DT_TEXT | uDrawTextAlignment);
						cur_rec.left -= 16;
						cur_rec.right -= sm_iSubItemInset;
						break;
					}

					case 9:
						if (client->GetUpPartCount()) {
							cur_rec.bottom--;
							cur_rec.top++;
							client->DrawUpStatusBar(dc, &cur_rec, false, thePrefs.UseFlatBar());
							cur_rec.bottom++;
							cur_rec.top--;
						}
						break;

					default:
						dc.DrawText(szItem, -1, &cur_rec, MLC_DT_TEXT | uDrawTextAlignment);
						break;
				}
			}
			cur_rec.left += iColumnWidth;
		}
	}

	DrawFocusRect(dc, lpDrawItemStruct->rcItem, lpDrawItemStruct->itemState & ODS_FOCUS, bCtrlFocused, lpDrawItemStruct->itemState & ODS_SELECTED);
}

void CQueueListCtrl::GetItemDisplayText(const CUpDownClient *client, int iSubItem, LPTSTR pszText, int cchTextMax)
{
	if (pszText == NULL || cchTextMax <= 0) {
		ASSERT(0);
		return;
	}
	pszText[0] = _T('\0');
	switch (iSubItem)
	{
		case 0:
			if (client->GetUserName() == NULL)
				_sntprintf(pszText, cchTextMax, _T("(%s)"), GetResString(IDS_UNKNOWN));
			else
				_tcsncpy(pszText, client->GetUserName(), cchTextMax);
			break;

		case 1: {
			const CKnownFile *file = theApp.sharedfiles->GetFileByID(client->GetUploadFileID());
			_tcsncpy(pszText, file != NULL ? file->GetFileName() : _T(""), cchTextMax);
			break;
		}

		case 2: {
			const CKnownFile *file = theApp.sharedfiles->GetFileByID(client->GetUploadFileID());
			if(file)
			{
			
				// <CB Mod : UploadFilePriority>
#ifndef CB_MOD_NEWPRIORITIES_ENABLED
				switch (file->GetUpPriority())
				{
					case PR_VERYLOW:
						_tcsncpy(pszText, GetResString(IDS_PRIOVERYLOW), cchTextMax);
						break;
					
					case PR_LOW:
						if (file->IsAutoUpPriority())
							_tcsncpy(pszText, GetResString(IDS_PRIOAUTOLOW), cchTextMax);
						else
							_tcsncpy(pszText, GetResString(IDS_PRIOLOW), cchTextMax);
						break;
					
					case PR_NORMAL:
						if (file->IsAutoUpPriority())
							_tcsncpy(pszText, GetResString(IDS_PRIOAUTONORMAL), cchTextMax);
						else
							_tcsncpy(pszText, GetResString(IDS_PRIONORMAL), cchTextMax);
						break;

					case PR_HIGH:
						if (file->IsAutoUpPriority())
							_tcsncpy(pszText, GetResString(IDS_PRIOAUTOHIGH), cchTextMax);
						else
							_tcsncpy(pszText, GetResString(IDS_PRIOHIGH), cchTextMax);
						break;

					case PR_VERYHIGH:
						_tcsncpy(pszText, GetResString(IDS_PRIORELEASE), cchTextMax);
						break;
				}

				if(file->GetPowerShared()) {
					_tcsncat(pszText, _T(" "), cchTextMax);
					_tcsncat(pszText, GetResString(IDS_POWERSHARE_PREFIX), cchTextMax);
				}
#else
				if (file->IsSuperRelease()) {
					_tcsncpy(pszText, GetResString(IDS_SUPERREL_PREFIX), cchTextMax);
				} else {
					_sntprintf(pszText, cchTextMax, _T("%i"), file->GetUpPriority());
					if (file->IsAutoUpPriority()) {
						_tcsncat(pszText, _T(" "), cchTextMax);
						_tcsncat(pszText, GetResString(IDS_PRIOAUTO), cchTextMax);
					}
					if (file->GetPowerShared()) {
						_tcsncat(pszText, _T(" "), cchTextMax);
						_tcsncat(pszText, GetResString(IDS_POWERSHARE_PREFIX), cchTextMax);
					}
				}
#endif
			}
			// </CB Mod : UploadFilePriority>

			break;
		}
		
		case 3: {
			// <CB Mod : Client Extended Info>
#ifdef CB_MOD_EXTRAINFO_ENABLED
			if (client->IsFriend() && client->GetFriendSlot()) {
				_tcsncpy(pszText, GetResString(IDS_FRIENDSLOT), cchTextMax);
			} else
#endif
			// </CB Mod : Client Extended Info>
				_sntprintf(pszText, cchTextMax, _T("%i"), client->GetScore(false, false, true));
		
					// <CB Mod : Client Extended Info>
#ifdef CB_MOD_EXTRAINFO_ENABLED
			sint64 diff = client->credits->GetUploadedTotal() - client->credits->GetDownloadedTotal();
			if (client->credits) {
				if (diff >= 0) {
					_tcsncat(pszText, _T("  [Diff: "), cchTextMax);
					_tcsncat(pszText, CastItoXBytes((uint64)diff, false, false), cchTextMax);
					_tcsncat(pszText, _T("]"), cchTextMax);
				} else {
					_tcsncat(pszText, _T("  [Diff: -"), cchTextMax);
					_tcsncat(pszText, CastItoXBytes((uint64)-diff, false, false), cchTextMax);
					_tcsncat(pszText, _T("]"), cchTextMax);
				}

#ifdef CB_MOD_BESTSHARE_ENABLED
				//if (client->credits->GetBonusTime() > 0) {
				if (theApp.uploadqueue->m_bestClient == client) {
					_tcsncat(pszText, _T(" /BS:"), cchTextMax);
					CString temp;
					temp.Format(_T("%u"), theApp.uploadqueue->GetBestShareMultiplier((int)-diff/PARTSIZE));
					_tcsncat(pszText, temp, cchTextMax);
				}
#endif
			}
#endif
				// </CB Mod : Client Extended Info>
		}
		break;

		case 4: {
				// <CB Mod : Client Extended Info>
#ifdef CB_MOD_ENHANCEDINTERFACE_ENABLED
				if (client->IsFriend() && client->GetFriendSlot()) {
					_sntprintf(pszText, cchTextMax, _T("%s %s"), GetResString(IDS_FRIENDSLOT), client->HasLowID()?L"("+GetResString(IDS_IDLOW)+L")":L"");
				} else {
#endif
				// </CB Mod : Client Extended Info>

					if (client->HasLowID()) {
						if (client->m_dwWouldHaveGottenUploadSlotIfNotLowIdTick)
							_sntprintf(pszText, cchTextMax, _T("%i **** Awaited reconnect %s"),client->GetScore(false), CastSecondsToHM((::GetTickCount()-client->m_dwWouldHaveGottenUploadSlotIfNotLowIdTick)/1000));
						else
							_sntprintf(pszText, cchTextMax, _T("%i (%s)"),client->GetScore(false), GetResString(IDS_IDLOW));
					}
					else
						_sntprintf(pszText, cchTextMax, _T("%i"), client->GetScore(false));
#ifdef CB_MOD_ENHANCEDINTERFACE_ENABLED
				}
#endif
			}
			break;

		case 5:
			_sntprintf(pszText, cchTextMax, _T("%i"), client->GetAskedCount());
			break;
		
		case 6:
			_tcsncpy(pszText, CastSecondsToHM((GetTickCount() - client->GetLastUpRequest()) / 1000), cchTextMax);
			break;

		case 7:
			
// <CB Mod : BestShare : Date 18.07.07>
#ifdef CB_MOD_BESTSHARE_ENABLED
			if (client && client->credits && client->credits->GetBonusTime() > 0)
				_sntprintf(pszText, cchTextMax, _T("%s (+ %s)"), CastSecondsToHM((::GetTickCount() - client->GetWaitStartTime())/1000), CastSecondsToHM(client->credits->GetBonusTime()/1000));
			else
				_tcsncpy(pszText, CastSecondsToHM((GetTickCount() - client->GetWaitStartTime()) / 1000), cchTextMax);
#else
			_tcsncpy(pszText, CastSecondsToHM((GetTickCount() - client->GetWaitStartTime()) / 1000), cchTextMax);
#endif
// </CB Mod : BestShare : Date 18.07.07>
			break;

		case 8:
			_tcsncpy(pszText, GetResString(client->IsBanned() ? IDS_YES : IDS_NO), cchTextMax);
			break;

		case 9:
			_tcsncpy(pszText, GetResString(IDS_UPSTATUS), cchTextMax);
			break;

        case 10:
			_tcsncpy(pszText, client->GetClientSoftVer(), cchTextMax);
            break;

        case 11:
			_tcsncpy(pszText, client->GetClientModVer(), cchTextMax);
            break;

		// <CB Mod : QueueListInfo>
#ifdef CB_MOD_EXTRAINFO_ENABLED
		case 12: {
				const CKnownFile *file = theApp.sharedfiles->GetFileByID(client->GetUploadFileID());
				if(file) {
					_tcsncpy(pszText, CastItoXBytes(file->GetFileSize(), false, false), cchTextMax);
				}
			}
			break;

		case 13:
			_tcsncpy(pszText, ipstr (client->GetIP()), cchTextMax);
			break;

		case 14:
			_sntprintf(pszText, cchTextMax, _T("%i"), client->m_QR);
#endif
		// </CB Mod : QueueListInfo>
	
	}
	pszText[cchTextMax - 1] = _T('\0');
}

void CQueueListCtrl::OnLvnGetDispInfo(NMHDR *pNMHDR, LRESULT *pResult)
{
	if (theApp.emuledlg->IsRunning()) {
		// Although we have an owner drawn listview control we store the text for the primary item in the listview, to be
		// capable of quick searching those items via the keyboard. Because our listview items may change their contents,
		// we do this via a text callback function. The listview control will send us the LVN_DISPINFO notification if
		// it needs to know the contents of the primary item.
		//
		// But, the listview control sends this notification all the time, even if we do not search for an item. At least
		// this notification is only sent for the visible items and not for all items in the list. Though, because this
		// function is invoked *very* often, do *NOT* put any time consuming code in here.
		//
		// Vista: That callback is used to get the strings for the label tips for the sub(!) items.
		//
		NMLVDISPINFO *pDispInfo = reinterpret_cast<NMLVDISPINFO*>(pNMHDR);
		if (pDispInfo->item.mask & LVIF_TEXT) {
			const CUpDownClient* pClient = reinterpret_cast<CUpDownClient*>(pDispInfo->item.lParam);
			if (pClient != NULL)
				GetItemDisplayText(pClient, pDispInfo->item.iSubItem, pDispInfo->item.pszText, pDispInfo->item.cchTextMax);
		}
	}
	*pResult = 0;
}

void CQueueListCtrl::OnLvnColumnClick(NMHDR *pNMHDR, LRESULT *pResult)
{
	NMLISTVIEW *pNMListView = (NMLISTVIEW *)pNMHDR;
	bool sortAscending;
	if (GetSortItem() != pNMListView->iSubItem)
	{
		switch (pNMListView->iSubItem)
		{
			case 2: // Up Priority
			case 3: // Rating
			case 4: // Score
			case 5: // Ask Count
			case 8: // Banned
			case 9: // Part Count
				sortAscending = false;
				break;
			default:
				sortAscending = true;
				break;
		}
	}
	else
		sortAscending = !GetSortAscending();

	// <CB Mod : 4-way sorting : Date 19.08.07>
#ifdef CB_MOD_EXTRAINFO_ENABLED
	int adder=0;
	int index = pNMListView->iSubItem - 3;
	if (pNMListView->iSubItem==3) {
		ASSERT( index < ARRSIZE(sortstat) );
		if (!sortAscending)
			sortstat[index] = !sortstat[index];
		adder = sortstat[index] ? 0 : 50;
	}

	UpdateSortHistory(pNMListView->iSubItem + adder + (sortAscending ? 0:100), 100);
	if (adder==0)	
		SetSortArrow(pNMListView->iSubItem, sortAscending); 
	else
		SetSortArrow(pNMListView->iSubItem, sortAscending ? arrowDoubleUp : arrowDoubleDown);
	SortItems(SortProc, pNMListView->iSubItem + adder + (sortAscending ? 0:100));

#else
	// Sort table
	UpdateSortHistory(pNMListView->iSubItem + (sortAscending ? 0 : 100));
	SetSortArrow(pNMListView->iSubItem, sortAscending);
	SortItems(SortProc, pNMListView->iSubItem + (sortAscending ? 0 : 100));
#endif
	// </CB Mod : 4-way sorting>

	*pResult = 0;
}

int CQueueListCtrl::SortProc(LPARAM lParam1, LPARAM lParam2, LPARAM lParamSort)
{
	const CUpDownClient *item1 = (CUpDownClient *)lParam1;
	const CUpDownClient *item2 = (CUpDownClient *)lParam2;
	int iColumn = (lParamSort >= 100) ? lParamSort - 100 : lParamSort;
	int iResult = 0;
	switch (iColumn)
	{
		case 0:
			if (item1->GetUserName() && item2->GetUserName())
				iResult = CompareLocaleStringNoCase(item1->GetUserName(), item2->GetUserName());
			else if (item1->GetUserName() == NULL)
				iResult = 1; // place clients with no usernames at bottom
			else if (item2->GetUserName() == NULL)
				iResult = -1; // place clients with no usernames at bottom
			break;

		case 1: {
			const CKnownFile *file1 = theApp.sharedfiles->GetFileByID(item1->GetUploadFileID());
			const CKnownFile *file2 = theApp.sharedfiles->GetFileByID(item2->GetUploadFileID());
			if (file1 != NULL && file2 != NULL)
				iResult = CompareLocaleStringNoCase(file1->GetFileName(), file2->GetFileName());
			else if (file1 == NULL)
				iResult = 1;
			else
				iResult = -1;
			break;
		}

		case 2: {
			const CKnownFile *file1 = theApp.sharedfiles->GetFileByID(item1->GetUploadFileID());
			const CKnownFile *file2 = theApp.sharedfiles->GetFileByID(item2->GetUploadFileID());
			if (file1 != NULL && file2 != NULL) {
				if(item1->GetPowerShared() == true && item2->GetPowerShared() == false)
                    iResult = 1;
                else if(item1->GetPowerShared() == false && item2->GetPowerShared() == true)
                    iResult = -1;
			    else
				// <CB Mod : UploadFilePriority>
#ifdef CB_MOD_NEWPRIORITIES_ENABLED
					iResult = file1->GetUpPriority() > file2->GetUpPriority() ? 1 : -1;
#else
				    iResult = (file1->GetUpPriority() == PR_VERYLOW ? -1 : file1->GetUpPriority()) - (file2->GetUpPriority() == PR_VERYLOW ? -1 : file2->GetUpPriority());
#endif
				// </CB Mod : UploadFilePriority>

			} else if (file1 == NULL)
				iResult = 1;
			else
				iResult = -1;
			break;
		}

		case 3: 
		// </CB Mod : Friend Management : Sorting : 03.02.08>
#ifdef CB_MOD_EXTRAINFO_ENABLED
			if (item1->GetFriendSlot() && !item2->GetFriendSlot()) {
				iResult = 1;
			} else if (!item1->GetFriendSlot() && item2->GetFriendSlot()) {
				iResult = -1;
			} else
			// Both or none have friendslot assigned
#endif
	// </CB Mod : Friend slot sorting>
			iResult=CompareUnsigned(item1->GetScore(false,false,true), item2->GetScore(false,false,true));
			break;
	
	// <CB Mod : 4-way sorting : Date 19.08.07>
#ifdef CB_MOD_EXTRAINFO_ENABLED
		case 53:
		{
			sint64 diff1 = item1->credits->GetUploadedTotal()-item1->credits->GetDownloadedTotal();
			sint64 diff2 = item2->credits->GetUploadedTotal()-item2->credits->GetDownloadedTotal();
			iResult = 0;
			if (item1->credits && item2->credits) {
				if (diff1 > diff2)
					iResult = 1;
				else if (diff1 < diff2)
					iResult = -1;
			}
		}
		break;
#endif
	// </CB Mod : 4-way sorting>

		case 4: {
				CKnownFile* file1 = theApp.sharedfiles->GetFileByID(item1->GetUploadFileID());
				CKnownFile* file2 = theApp.sharedfiles->GetFileByID(item2->GetUploadFileID());
				if( (file1 != NULL) && (file2 != NULL)) {
					if(item1->GetPowerShared() == true && item2->GetPowerShared() == false)
						iResult = 1;
					else if(item1->GetPowerShared() == false && item2->GetPowerShared() == true)
						iResult = -1;
					else if(item1->GetPowerShared() == true && item2->GetPowerShared() == true)
					// <CB Mod : UploadFilePriority>
#ifdef CB_MOD_NEWPRIORITIES_ENABLED
						iResult = file1->GetUpPriority() > file2->GetUpPriority() ? 1 : -1;
#else
						iResult = ((file1->GetUpPriority()==PR_VERYLOW) ? -1 : file1->GetUpPriority()) - ((file2->GetUpPriority()==PR_VERYLOW) ? -1 : file2->GetUpPriority());
#endif
					// </CB Mod : UploadFilePriority>
					else
						iResult = 0;
			} else if( file1 == NULL )
				iResult = 1;
			else
				iResult = -1;

			if(iResult == 0) {
				iResult=CompareUnsigned(item1->GetScore(false), item2->GetScore(false));
				}
			}
			break;

		case 5:
			iResult = CompareUnsigned(item1->GetAskedCount(), item2->GetAskedCount());
			break;

		case 6:
			iResult = CompareUnsigned(item1->GetLastUpRequest(), item2->GetLastUpRequest());
			break;

		case 7:
			iResult = CompareUnsigned(item1->GetWaitStartTime(), item2->GetWaitStartTime());
			break;

		case 8:
			iResult = item1->IsBanned() - item2->IsBanned();
			break;
		
		case 9: 
			iResult = CompareUnsigned(item1->GetUpPartCount(), item2->GetUpPartCount());
			break;

        case 10:
		// <CB Mod : Fix>
			if(!item1->GetClientSoftVer().IsEmpty() && !item2->GetClientSoftVer().IsEmpty()) {
				iResult = _tcsicmp(item1->GetClientSoftVer(), item2->GetClientSoftVer());
		    } else if(item1->GetClientSoftVer())
			    iResult =  1;
		    else
			    iResult =  -1;

            break;

        case 11:
		  if(!item1->GetClientModVer().IsEmpty() && !item2->GetClientModVer().IsEmpty()) {
				iResult = _tcsicmp(item1->GetClientModVer(), item2->GetClientModVer());

		  } else
				if (item1->GetClientModVer().IsEmpty())
					iResult =  1;
				else if (item2->GetClientModVer().IsEmpty())
					iResult =  -1;
				else
					iResult = 0; // keep peers sorted as for latest sort critera
		// </CB Mod : Fix>

            break;

		// <CB Mod : QueueListInfo : Sort>
#ifdef CB_MOD_EXTRAINFO_ENABLED
		case 12: // File size ascending
		{
			CKnownFile *file1 = theApp.sharedfiles->GetFileByID(item1->GetUploadFileID());
			CKnownFile *file2 = theApp.sharedfiles->GetFileByID(item2->GetUploadFileID());
			if( (file1 != NULL) && (file2 != NULL))
				iResult = CompareUnsigned64((uint64)file1->GetFileSize(), (uint64)file2->GetFileSize());
		}

		case 13: // IP Number ascending
		{
			uint32 ip1 = ntohl(item1->GetIP());
			uint32 ip2 = ntohl(item2->GetIP());
			if(ip1 > ip2)
				iResult = -1;
			else if (ip1 < ip2)
				iResult = 1;
			else
				iResult = 0;
		}

		case 14: // QR value ascending
		{
			iResult = item1->m_QR - item2->m_QR;
		}
#endif
		// </CB Mod : QueueListInfo : Sort>

	}

	if (lParamSort >= 100)
		iResult = -iResult;

	//call secondary sortorder, if this one results in equal
	int dwNextSort;
	if (iResult == 0 && (dwNextSort = theApp.emuledlg->transferwnd->queuelistctrl.GetNextSortOrder(lParamSort)) != -1)
		iResult = SortProc(lParam1, lParam2, dwNextSort);

	return iResult;
}

void CQueueListCtrl::OnNmDblClk(NMHDR* /*pNMHDR*/, LRESULT *pResult)
{
	int iSel = GetNextItem(-1, LVIS_SELECTED | LVIS_FOCUSED);
	if (iSel != -1) {
		CUpDownClient* client = (CUpDownClient*)GetItemData(iSel);
		if (client){
			CClientDetailDialog dialog(client, this);
			dialog.DoModal();
		}
	}
	*pResult = 0;
}

void CQueueListCtrl::OnContextMenu(CWnd* /*pWnd*/, CPoint point)
{
	int iSel = GetNextItem(-1, LVIS_SELECTED | LVIS_FOCUSED);
	const CUpDownClient* client = (iSel != -1) ? (CUpDownClient*)GetItemData(iSel) : NULL;

	CTitleMenu ClientMenu;
	ClientMenu.CreatePopupMenu();
	ClientMenu.AddMenuTitle(GetResString(IDS_CLIENTS), true);
	ClientMenu.AppendMenu(MF_STRING | (client ? MF_ENABLED : MF_GRAYED), MP_DETAIL, GetResString(IDS_SHOWDETAILS), _T("CLIENTDETAILS"));
	ClientMenu.SetDefaultItem(MP_DETAIL);
	ClientMenu.AppendMenu(MF_STRING | ((client && client->IsEd2kClient() && !client->IsFriend()) ? MF_ENABLED : MF_GRAYED), MP_ADDFRIEND, GetResString(IDS_ADDFRIEND), _T("ADDFRIEND"));
		
	// <CB Mod : Friend Management : Easy Remove>
#ifdef CB_MOD_EASYFRIENDSLOT_ENABLED
	ClientMenu.AppendMenu(MF_STRING | (client && client->IsFriend() ? MF_ENABLED : MF_GRAYED), MP_REMOVEFRIEND, GetResString(IDS_REMOVEFRIEND), _T("DELETEFRIEND"));
#endif
	// </CB Mod : Friend Management : Easy Remove>
	
	ClientMenu.AppendMenu(MF_STRING | ((client && client->IsEd2kClient()) ? MF_ENABLED : MF_GRAYED), MP_MESSAGE, GetResString(IDS_SEND_MSG), _T("SENDMESSAGE"));
	ClientMenu.AppendMenu(MF_STRING | ((client && client->IsEd2kClient() && client->GetViewSharedFilesSupport()) ? MF_ENABLED : MF_GRAYED), MP_SHOWLIST, GetResString(IDS_VIEWFILES), _T("VIEWFILES"));
	if (thePrefs.IsExtControlsEnabled())
		ClientMenu.AppendMenu(MF_STRING | ((client && client->IsEd2kClient() && client->IsBanned()) ? MF_ENABLED : MF_GRAYED), MP_UNBAN, GetResString(IDS_UNBAN));
	if (Kademlia::CKademlia::IsRunning() && !Kademlia::CKademlia::IsConnected())
		ClientMenu.AppendMenu(MF_STRING | ((client && client->IsEd2kClient() && client->GetKadPort()!=0) ? MF_ENABLED : MF_GRAYED), MP_BOOT, GetResString(IDS_BOOTSTRAP));
	ClientMenu.AppendMenu(MF_STRING | (GetItemCount() > 0 ? MF_ENABLED : MF_GRAYED), MP_FIND, GetResString(IDS_FIND), _T("Search"));

	// <CB Mod : Friend Management : Easy FriendSlot>
#ifdef CB_MOD_EASYFRIENDSLOT_ENABLED
	ClientMenu.AppendMenu(MF_SEPARATOR);
	ClientMenu.AppendMenu(MF_STRING, MP_FRIENDSLOT, GetResString(IDS_FRIENDSLOT), _T("FRIENDSLOT"));
	ClientMenu.EnableMenuItem(MP_FRIENDSLOT, (client)?MF_ENABLED:MF_GRAYED);
	ClientMenu.CheckMenuItem(MP_FRIENDSLOT, (client && client->GetFriendSlot())? MF_CHECKED : MF_UNCHECKED);
#endif
	// <CB Mod : Friend Management : Easy FriendSlot>

	// <CB Mod : ForceCheck>
#if defined(CB_MOD_ILLEGALMODSFILTER_ENABLED) && defined(_DEBUG)
	ClientMenu.AppendMenu(MF_SEPARATOR);
	ClientMenu.AppendMenu(MF_STRING, MP_FORCECLIENTCHECK, _T("Force Check"), NULL);
#endif
	// <CB Mod : ForceCheck>

	GetPopupMenuPos(*this, point);
	ClientMenu.TrackPopupMenu(TPM_LEFTALIGN |TPM_RIGHTBUTTON, point.x, point.y, this);
}

BOOL CQueueListCtrl::OnCommand(WPARAM wParam, LPARAM /*lParam*/)
{
	wParam = LOWORD(wParam);

	switch (wParam)
	{
		case MP_FIND:
			OnFindStart();
			return TRUE;
	}

	int iSel = GetNextItem(-1, LVIS_SELECTED | LVIS_FOCUSED);
	if (iSel != -1){
		CUpDownClient* client = (CUpDownClient*)GetItemData(iSel);
	
		// <CB Mod : ContextMenu>
#ifdef CB_MOD_GENERALFIXES_ENABLED
		// It may happen that the client is removed from the queue when the a selection has been done
		if (!client)
			return true;
#endif
		// </CB Mod : ContextMenu>

		switch (wParam){
			case MP_SHOWLIST:
				client->RequestSharedFileList();
				break;
			case MP_MESSAGE:
				theApp.emuledlg->chatwnd->StartSession(client);
				break;
			case MP_ADDFRIEND:
				if (theApp.friendlist->AddFriend(client))
					Update(iSel);
				break;
			case MP_UNBAN:
				if (client->IsBanned()){
					client->UnBan();
					Update(iSel);
				}
				break;
			case MP_DETAIL:
			case MPG_ALTENTER:
			case IDA_ENTER:
			{
				CClientDetailDialog dialog(client, this);
				dialog.DoModal();
				break;
			}
			case MP_BOOT:
				if (client->GetKadPort())
					Kademlia::CKademlia::Bootstrap(ntohl(client->GetIP()), client->GetKadPort(), (client->GetKadVersion() > 1));
				break;

			// <CB Mod : Friend Management : Easy FriendSlot>
#ifdef CB_MOD_EASYFRIENDSLOT_ENABLED
			case MP_FRIENDSLOT:
				theApp.friendlist->AddFriend(client);
				if (client->GetFriendSlot())
					client->SetFriendSlot(false);
				else
					client->SetFriendSlot(true);
				Update(iSel);
				break;

			case MP_REMOVEFRIEND:
				if (client && client->IsFriend() && client->m_Friend != NULL) {
					theApp.friendlist->RemoveFriend(client->m_Friend);
					Update(iSel);	
				}
				break;
#endif
			// </CB Mod : Friend Management : Easy FriendSlot>


			// <CB Mod : ForceCheck>
#if defined(CB_MOD_ILLEGALMODSFILTER_ENABLED) && defined(_DEBUG)
			case MP_FORCECLIENTCHECK:
				client->CheckForGPLEvilDoer();
				break;
#endif
			// </CB Mod : ForceCheck>
		}
	}
	return true;
}

void CQueueListCtrl::AddClient(/*const*/ CUpDownClient *client, bool resetclient)
{
	if (resetclient && client){
		client->SetWaitStartTime();
		client->SetAskedCount(1);
    } else if( client ) {
        // Clients that have been put back "first" on queue (that is, they
        // get to keep its waiting time since before they started upload), are
        // recognized by having an ask count of 0.
        client->SetAskedCount(0);
	}

	if (!theApp.emuledlg->IsRunning())
		return;
	if (thePrefs.IsQueueListDisabled())
		return;

	int iItemCount = GetItemCount();
	
	// <CB Mod : QueueListInfo : QR>
#ifdef CB_MOD_EXTRAINFO_ENABLED
	client->m_QR = iItemCount + 1;
#endif
	// </CB Mod : QueueListInfo : QR>

	int iItem = InsertItem(LVIF_TEXT | LVIF_PARAM, iItemCount, LPSTR_TEXTCALLBACK, 0, 0, 0, (LPARAM)client);
	Update(iItem);
	theApp.emuledlg->transferwnd->UpdateListCount(CTransferWnd::wnd2OnQueue, iItemCount + 1);
}

void CQueueListCtrl::RemoveClient(const CUpDownClient *client)
{
	if (!theApp.emuledlg->IsRunning())
		return;

	LVFINDINFO find;
	find.flags = LVFI_PARAM;
	find.lParam = (LPARAM)client;
	int result = FindItem(&find);
	if (result != -1) {
		DeleteItem(result);
		theApp.emuledlg->transferwnd->UpdateListCount(CTransferWnd::wnd2OnQueue);
	}
}

void CQueueListCtrl::RefreshClient(const CUpDownClient *client)
{
	if (!theApp.emuledlg->IsRunning())
		return;

	if (theApp.emuledlg->activewnd != theApp.emuledlg->transferwnd || !theApp.emuledlg->transferwnd->queuelistctrl.IsWindowVisible())
		return;

	LVFINDINFO find;
	find.flags = LVFI_PARAM;
	find.lParam = (LPARAM)client;
	int result = FindItem(&find);
	if (result != -1)
		Update(result);
}

void CQueueListCtrl::ShowSelectedUserDetails()
{
	POINT point;
	::GetCursorPos(&point);
	CPoint p = point; 
    ScreenToClient(&p); 
    int it = HitTest(p); 
    if (it == -1)
		return;

	SetItemState(-1, 0, LVIS_SELECTED);
	SetItemState(it, LVIS_SELECTED | LVIS_FOCUSED, LVIS_SELECTED | LVIS_FOCUSED);
	SetSelectionMark(it);   // display selection mark correctly!

	CUpDownClient* client = (CUpDownClient*)GetItemData(GetSelectionMark());
	if (client){
		CClientDetailDialog dialog(client, this);
		dialog.DoModal();
	}
}

void CQueueListCtrl::ShowQueueClients()
{
	DeleteAllItems(); 
	CUpDownClient *update = theApp.uploadqueue->GetNextClient(NULL);
	while( update )
	{
		AddClient(update, false);
		update = theApp.uploadqueue->GetNextClient(update);
	}
}

// Barry - Refresh the queue every 10 secs
void CALLBACK CQueueListCtrl::QueueUpdateTimer(HWND /*hwnd*/, UINT /*uiMsg*/, UINT /*idEvent*/, DWORD /*dwTime*/)
{
	// NOTE: Always handle all type of MFC exceptions in TimerProcs - otherwise we'll get mem leaks
	try
	{
		if (   !theApp.emuledlg->IsRunning() // Don't do anything if the app is shutting down - can cause unhandled exceptions
			|| !thePrefs.GetUpdateQueueList()
			|| theApp.emuledlg->activewnd != theApp.emuledlg->transferwnd
			|| !theApp.emuledlg->transferwnd->queuelistctrl.IsWindowVisible() )
			return;

		const CUpDownClient* update = theApp.uploadqueue->GetNextClient(NULL);
		while( update )
		{
			theApp.emuledlg->transferwnd->queuelistctrl.RefreshClient(update);
			update = theApp.uploadqueue->GetNextClient(update);
		}
	}
	CATCH_DFLT_EXCEPTIONS(_T("CQueueListCtrl::QueueUpdateTimer"))
}

